/* b016 --- device driver for the INMOS B016                26/01/1990 */
/* Copyright (c) 1990 John Honniball, INMOS Ltd.                       */

/* Modification:
   26/01/90 BJ  Initial coding
*/

#include <sys/param.h>
#include <sys/types.h>
#include <sys/buf.h>
#include <sys/user.h>
#include <machine/psl.h>
#include <machine/param.h>
#include <sundev/mbvar.h>     /* definition of 'mb_driver' */

#include <sys/ims_bcmd.h>
#include "bxvi.h"

#include "b016reg.h"


/* #define DB*/
/* #define DB1 */    /* Installation debugging */
/* #define DB2 */    /* Dump state on open & close */

#define YES       1
#define NO        0

#define READ_CHAN    0
#define WRITE_CHAN   1

#define BIT_0        1
#define BIT_1        2
#define BIT_2        4
#define BIT_3        8

#define N_LINKS         4                       /* Links per board   */
#define MAX_BOARDS      8                       /* Boards per system */
#define MAX_LINKS       (MAX_BOARDS * N_LINKS)  /* Links per system  */

#define B016UNIT(dev)   (minor (dev))
#define B016CTLR(dev)   ((minor (dev)) / N_LINKS)
#define B016LINK(dev)   ((minor (dev)) % N_LINKS)

#define B016PRI         (PZERO - 1)

#define MAX_BUF         32768
#define MEG             (1024 * 1024)

#define DEFAULT_TIMEOUT 1000           /* 100 seconds */

#define B016_BLOCK_SIZE  32768         /* Size of B016 shared RAM buffer */
#define DP_MEM_SIZE     (2 * MEG)

typedef unsigned char BYTE;
typedef unsigned int BOOL;

/* Data structures that reflect the state of the hardware */

struct ChanState {
   int ErrCode;                     /* Error code from int. routine    */
   struct buf Buf;                  /* Buffer for 'physio'             */
   int Timeout;                     /* Channel timeout (tenths of sec) */
   char *B016Buf;                   /* Shared RAM buffer address       */
   long int *B016Len;               /* Shared RAM length word address  */
   long int *B016Time;              /* Shared RAM timeout word address */
};

struct LinkState {
   int ErrCode;                     /* Error code from int. routine    */
   BOOL ErrFlag;                    /* State of transputer's error bit */
   BOOL Open;                       /* TRUE, if device is open already */
   BOOL Timed;                      /* TRUE, if watchdog timer needed  */
   BOOL Signal;                     /* Signal reqd for error           */
   int Signo;                       /* Signal level for use in error   */
   struct proc *UsrProc;            /* Pointer to the per user process */
   struct ChanState Chan[2];        /* Two channels (read and write)   */
};

struct BoardState {
   BOOL Busy;                       /* Busy flag for mutual exclusion  */
   caddr_t MemAddr;                 /* Physical address of shared RAM  */
   char *SharedRam;                 /* Virtual address of shared RAM   */
   struct B016Reg *Reg;             /* Virtual address of VIC regs     */
   struct LinkState Link[N_LINKS];  /* Four links                      */
};

struct BoardState B016Board[MAX_BOARDS];


/* Kernel interface for auto-configuration */

int B016VicProbe (), B016Attach (), bxviIntr ();
int B016Slave ();

struct mb_ctlr   *B016Ctlr[MAX_BOARDS];
struct mb_device *B016Info[MAX_LINKS];

struct mb_driver bxviVicdriver = {
   B016VicProbe,              /* Probe routine              */
   B016Slave,                 /* Slaves                     */
   B016Attach,                /* Attach routine             */
   0,                         /* Go routine                 */
   0,                         /* Done routine               */
   bxviIntr,                  /* Interrupt routine          */
   sizeof (struct B016Reg),   /* Memory space               */
   "bxvi",                    /* Device name                */
   B016Info,                  /* Ptrs to mbdinit structs    */
   "B016Vic",                 /* Controller name            */
   B016Ctlr,                  /* Ptrs to mbcinit structs    */
   0,                         /* Exclusive use of Main Bus  */
   0                          /* Interrupt routine list     */
};

int B016MemProbe ();

struct mb_driver bxviMemdriver = {
   B016MemProbe,              /* Probe routine              */
   0,                         /* Slaves                     */
   0,                         /* Attach routine             */
   0,                         /* Go routine                 */
   0,                         /* Done routine               */
   0,                         /* Interrupt routine          */
   DP_MEM_SIZE,               /* Dual-port memory space     */
   "B016Mem",                 /* Device name                */
   0,                         /* Ptrs to mbdinit structs    */
   "B016Mem",                 /* Controller name            */
   B016Ctlr,                  /* Ptrs to mbcinit structs    */
   0,                         /* Exclusive use of Main Bus  */
   0                          /* Interrupt routine list     */
};

/*
 * The following variables implement communication between the 'B016Mem'
 * driver and the 'B016Vic' driver.  The 'B016Mem' driver allocates two
 * megabytes of memory for the B016's shared RAM.  The 'B016Vic' driver
 * allocates a few bytes for the VIC chip.  The virtual and physical
 * addresses of the shared RAM are passed across to the VIC driver during
 * probing.  See the Sun 'Writing Device Drivers' manual, Section 7.3,
 * page 145 for an explanation of the enormity of this kludge.
 */
static caddr_t MemAddr = 0;
static char *SharedRam = NULL;

extern int hz;    /* Kernel clock rate */

static int MinPhys (/* struct buf *bp */);
static int Strategy (/* struct buf *bp */);
static int ReadFlags (/* caddr_t data, struct LinkState *lp */);
static int SetFlags (/* caddr_t data, int ctlr, int link */);
static int SendCmd (/* int ctlr, int link, int cmd */);
static void InitLink (/* int ctlr, int link */);
static void InitChan (/* int ctlr, int link, int chan */);
static void DumpState (/* int ctlr, int link */);


/* B016MemProbe --- stash memory address */

int B016MemProbe (reg, unit)
caddr_t reg;      /* Address of the device */
int unit;         /* Device number         */
{
/*
 * This is the only routine in the 'stub' driver.  All it does is accept
 * the address 'reg' from the kernel and stash it in 'SharedRam' and, as
 * a physical address, in 'MemAddr'.
 */
   int pf;     /* Page Frame index */
   
   if (MemAddr == 0) {
      pf = hat_getkpfnum (reg);
      MemAddr = (caddr_t)(ptob (pf));
      SharedRam = reg;

#ifdef DB1
      printf ("MemProbe: reg = 0x%x, unit = %d\n", reg, unit);
      printf ("Physical: 0x%x\n", MemAddr);
#endif
   
      /* We can't actually probe this yet, 'cos the B016 isn't ready */
      return (DP_MEM_SIZE);
   }
   else {
      printf ("B016 device drivers installed in wrong order\n");
      return (0);
   }
}


/* B016VicProbe --- detect presence of B016 */

int B016VicProbe (reg, ctlr)
caddr_t reg;      /* Address of the device */
int ctlr;         /* Controller number     */
{
/*
 * This routine is supposed to merely detect the presence of a B016 and
 * return.   It also initialises an element of the 'B016Board' array.
 */
   int i;
   int c;
   struct B016Reg *r;
   
#ifdef DB1
   printf ("VicProbe: reg = 0x%x, ctlr = %d\n", reg, ctlr);
#endif

   r = (struct B016Reg *)reg;
   
   /* Make sure all five mailboxes exist and contain 42 */
   for (i = 0; i < 5; i++) {
      c = peekc (&(r->vic[i].reg));
      if (c == -1) {
#ifdef DB1
         printf ("Vic not found - bus error\n");
#endif

         return (0);    /* No device installed */
      }
      else if (c != 42) {
         printf ("Vic found OK - not booted\n");

         return (0);    /* B016 didn't boot up */
      }
   }
   
   /* Ensure the other 'stub' driver got there first */
   if (MemAddr  == 0) {
      printf ("B016 device drivers installed in wrong order\n");
      return (0);
   }
   
   B016Board[ctlr].Busy      = NO;
   B016Board[ctlr].Reg       = r;
   B016Board[ctlr].MemAddr   = MemAddr;
   B016Board[ctlr].SharedRam = SharedRam;

#ifdef DB1
   printf ("VicProbe: returns %d\n", sizeof (struct B016Reg));
#endif

   return (sizeof (struct B016Reg));   /* Found a B016 */
}


/* B016Slave --- detect presence of slaves */

int B016Slave (md, reg)
struct mb_device *md;
caddr_t reg;      /* Address of the device */
{
/*
 * This routine is not documented in the Sun 'Writing Device Drivers'
 * manual.  It is mentioned on page 53 in connection with 'mb_driver'
 * structures but the details of arguments and return value are not
 * specified.  A call to Sun technical support revealed the above
 * 'md' and 'reg' arguments.
 */
#ifdef DB1
   printf ("Slave: reg = 0x%x, slave = %d\n", reg, md->md_slave);
#endif

   /* Check number of links per board */
   if (md->md_slave >= N_LINKS)
      return (0);
      
   return (sizeof (struct B016Reg));
}


/* B016Attach --- initialise the B016 slaves */

int B016Attach (md)
struct mb_device *md;   /* Pointer to slave device structure */
{
/*
 * There's no work to be done to initialise the hardware because the
 * T801 on the B016 has done it all already.  We just need to set up
 * the per-link data structure 'B016Link[]'.
 *
 * In fact, since there is no controller-level 'Attach ()'
 * routine (only slave-level), this routine also initialises the B016.
 * This is only done when 'link' is zero, i.e. for the first link
 * on each B016.
 * The main task is to send parameters through the VIC mailboxes.
 */
   int unit;               /* Unit number                */
   int link;               /* Link number                */
   int ctlr;               /* Controller number          */
   struct mb_ctlr *mc;     /* We need the interrupt info */
   struct B016Reg *r;      /* Pointer to B016 registers  */
   int addr;               /* Address of shared memory   */
   
   unit = md->md_unit;
   link = md->md_slave;
   ctlr = unit / N_LINKS;

   r  = B016Board[ctlr].Reg;
   mc = B016Ctlr[ctlr];

#ifdef DB1
   printf ("Attach: unit = %d, link = %d, ctlr = %d\n", unit, link, ctlr);
#endif

   /* Should we initialise the controller ? */
   if (link == 0) {
      /* Get address of shared memory */
      addr = (int)B016Board[ctlr].MemAddr;
      
      /* Send base address to B016 via mailboxes */
      r->vic[0].reg =  addr        & 0xff;
      r->vic[1].reg = (addr >> 8)  & 0xff;
      r->vic[2].reg = (addr >> 16) & 0xff;
      r->vic[3].reg = (addr >> 24) & 0xff;
      r->vic[4].reg = 0xff;

      /* Wait for B016 to acknowledge base address */
      while (r->vic[4].reg == 0xff)
         ;

      /* Send the interrupt level and vector */
      if (mc->mc_intr != NULL) {
#ifdef DB1
         printf ("Attach: vector = 0x%x, level = %d\n",
                             mc->mc_intr->v_vec, mc->mc_intpri);
#endif
         r->vic[0].reg = mc->mc_intr->v_vec;
         r->vic[1].reg = mc->mc_intpri;
      }
      else {
         printf ("B016: no interrupt vector specified in config\n");
         r->vic[0].reg = 0;
         r->vic[1].reg = 0;
      }

      r->vic[4].reg = 0x02; /* Signal the interrupt info is ready */
      
      /* Wait for B016 to acknowledge interrupt info */
      while (r->vic[4].reg == 0x02)
         ;

#ifdef DB1
      printf ("Attach: B016 accepted address\n");
#endif
   }

   /* Indicate that we've used the value of 'MemAddr' */
   MemAddr = 0;
   
   /* Initialise all fields in the state structure */
   InitLink (ctlr, link);
}


/* B016Open --- open communications with the device */

int B016Open (dev, flags)
dev_t dev;     /* Major and minor device numbers */
int flags;     /* Flags passed from user         */
{
   int unit;
   int ctlr;
   int link;
   struct mb_device *md;
   struct mb_ctlr   *mc;
   struct LinkState *lp;
   
   unit = B016UNIT (dev);
   ctlr = B016CTLR (dev);
   link = B016LINK (dev);
   
#ifdef DB
   printf ("Open: unit = %d, flags = %d\n", unit, flags);
#endif

   if (unit >= MAX_LINKS)     /* Validate unit number */
      return (ENXIO);
      
   md = B016Info[unit];             /* Get mb_device pointer */
   mc = B016Ctlr[ctlr];             /* Get mb_ctlr pointer   */

   if (mc == NULL || md == NULL)
      return (ENXIO);         /* No device */
   
   if (mc->mc_alive == NO || md->md_alive == NO)
      return (ENXIO);         /* No device */
   
   lp = &B016Board[ctlr].Link[link];

   /* Can't be opened if it's already open */
   if (lp->Open)
      return (EBUSY);         /* Already open */

   lp->Open = YES;

#ifdef DB2
   DumpState (ctlr, link);
#endif

   return (0);             /* Success code */
}


/* MinPhys --- adjust 'b_bcount' for max block size */

static int MinPhys (bp)
register struct buf *bp;
{
   if (bp->b_bcount > B016_BLOCK_SIZE)
      bp->b_bcount = B016_BLOCK_SIZE;
}


/* Strategy --- high-level I/O routine */

static int Strategy (bp)
register struct buf *bp;
{
   int ctlr;
   int link;
   struct mb_ctlr *mc;
   struct ChanState *cp;
   int s;
   int ret;
   
   if (bp->b_flags & B_READ)
      bp_mapin (bp);       /* Lock those pages down */

#ifdef DB
   printf ("Strat: %d bytes to %s\n",
                   bp->b_bcount, (bp->b_flags & B_READ) ? "read" : "write");
#endif

   ctlr = B016CTLR (bp->b_dev);
   link = B016LINK (bp->b_dev);

   mc = B016Ctlr[ctlr];      /* Get mb_ctlr pointer   */

   /*
    * Lock the device structure in order to implement mutual
    * exclusion.  Note that the interrupt routine uses different
    * registers in the VIC and therefore does not need to be
    * locked in this way.
    */
   s = splx (pritospl (mc->mc_intpri));   /* +++ begin critical section +++ */
   while (B016Board[ctlr].Busy)
      sleep ((caddr_t)&B016Board[ctlr], B016PRI);

   B016Board[ctlr].Busy = YES;

   /* Are we reading or writing ? */
   if (bp->b_flags & B_READ) {
      cp = &B016Board[ctlr].Link[link].Chan[READ_CHAN];
      
      /* Tell B016 length of packet and timeout duration */
      *(cp->B016Len) = bp->b_bcount;
      *(cp->B016Time) = cp->Timeout;

      if (SendCmd (ctlr, link, READ_LINK_CMD) == 0) {
         bp->b_flags |= B_ERROR;
         bp->b_error = EIO;
         iodone (bp);      /* Pretend that the I/O was actually done */
      }
   }
   else {   /* Must be a write... */
      cp = &B016Board[ctlr].Link[link].Chan[WRITE_CHAN];

      /* Transfer data block to B016 */
      *(cp->B016Len) = bp->b_bcount;
      *(cp->B016Time) = cp->Timeout;
      bcopy (bp->b_un.b_addr, cp->B016Buf, (int)bp->b_bcount);

      if (SendCmd (ctlr, link, WRITE_LINK_CMD) == 0) {
         bp->b_flags |= B_ERROR;
         bp->b_error = EIO;
         iodone (bp);      /* Pretend that the I/O was actually done */
      }
   }
   
   /*
    * Awaken any other 'strategy ()' routines that are waiting for
    * access to the B016.
    */
   B016Board[ctlr].Busy = NO;    /* Board is no longer busy */
   wakeup ((caddr_t)&B016Board[ctlr]);
   (void)splx (s);                        /* --- end critical section --- */
   
#ifdef DB
   printf ("Strat: exit\n");
#endif
}


/* B016Read --- read data from the B016 */

int B016Read (dev, uio)
dev_t dev;              /* Major and minor device numbers */
struct uio *uio;        /* */
{
   int ctlr;
   int link;
   int ret;
   register struct buf *bp;
   
   ctlr = B016CTLR (dev);
   link = B016LINK (dev);
   
#ifdef DB
   printf ("Read: ctlr %d, link %d\n", ctlr, link);
#endif
   
   if (ctlr >= MAX_BOARDS)    /* Validate controller number */
      return (ENXIO);

   /* Get buffer pointer */
   bp = &B016Board[ctlr].Link[link].Chan[READ_CHAN].Buf;
   
   ret = physio (Strategy, bp, dev, B_READ, MinPhys, uio);
   
#ifdef DB
   printf ("Read: physio returned %d\n", ret);
#endif

   return (ret);
}


/* B016Write --- write data into the B016 */

int B016Write (dev, uio)
dev_t dev;              /* Major and minor device numbers */
struct uio *uio;        /* */
{
   int ctlr;
   int link;
   int ret;
   register struct buf *bp;
   
   ctlr = B016CTLR (dev);
   link = B016LINK (dev);
   
#ifdef DB
   printf ("Write: ctlr %d, link %d\n", ctlr, link);
#endif
   
   if (ctlr >= MAX_BOARDS)    /* Validate controller number */
      return (ENXIO);

   /* Get buffer pointer */
   bp = &B016Board[ctlr].Link[link].Chan[WRITE_CHAN].Buf;
   
   ret = physio (Strategy, bp, dev, B_WRITE, MinPhys, uio);

#ifdef DB
   printf ("Write: physio returned %d\n", ret);
#endif
   
   return (ret);
}


/* B016Close --- close down the device */

int B016Close (dev, flags)
dev_t dev;        /* Major and minor device numbers */
int flags;        /* Read/write access flags        */
{
   int link;
   int ctlr;
   
   ctlr = B016CTLR (dev);
   link = B016LINK (dev);
   
#ifdef DB
   printf ("Close: ctlr %d, link %d\n", ctlr, link);
#endif
   
   if (ctlr >= MAX_BOARDS)    /* Validate controller number */
      return (ENXIO);
   
   /* Restore initial state of link data structure */
   InitLink (ctlr, link);
   
#ifdef DB2
   DumpState (ctlr, link);
#endif
   
   return (0);                /* Success code */
}


/* bxviIntr --- interrupt handler */

int bxviIntr (ctlr)
int ctlr;      /* Controller number */
{
   struct B016Reg *r;
   struct LinkState *lp;
   struct ChanState *cp;
   struct buf *bp;
   int link;               /* Link number from B016 register      */
   int reason;             /* Interrupt reason from B016 register */
   int errcode;            /* Error code from B016 register       */
   
   /* Read stuff from VIC registers */
   r = B016Board[ctlr].Reg;

   link    = r->vic[1].reg;
   reason  = r->vic[2].reg;
   errcode = r->vic[3].reg;

#ifdef DB
   printf ("B016Intr: ctlr %d, link %d, errcode %d: ", ctlr, link, errcode);
#endif
   
   if (link < 0 || link >= N_LINKS) {
      printf ("B016Vic%d: invalid link: %d\n", ctlr, link);
      return (1);
   }
   
   lp = &B016Board[ctlr].Link[link];
   
#ifdef DB
   switch (reason) {
   case 0:
      printf ("IOCTL\n");
      break;
   case 1:
      printf ("ERROR\n");
      break;
   case 2:
      printf ("READ\n");
      break;
   case 3:
      printf ("WRITE\n");
      break;
   }
#endif

   /* Decide what we're doing */
   switch (reason) {
   case 0:
      lp->ErrCode = errcode;  /* Store the error for 'ReadFlags' to pick up */
      wakeup ((caddr_t)lp);   /* Awaken the waiting IOCTL code              */
      break;
      
   case 1:
      if (errcode == E_SETERROR)
         lp->ErrFlag = YES;
      else if (errcode == E_CLRERROR)
         lp->ErrFlag = NO;
      else
         printf ("B016Vic%d: link %d error flag confused\n", ctlr, link);
         
      break;
      
   case 2:
   case 3:
      if (reason == 2)
         cp = &(lp->Chan[READ_CHAN]);
      else
         cp = &(lp->Chan[WRITE_CHAN]);

      bp = &(cp->Buf);
      
      cp->ErrCode = errcode;
      
      /* Check for error codes from B016 */
      if (errcode == E_TIMEOUT) {
         bp->b_flags |= B_ERROR;
         bp->b_error = ETIME;
      }
      else if (bp->b_flags & B_READ) {
         /* If reading, copy data from B016 into buffer */
         bcopy (cp->B016Buf, bp->b_un.b_addr, (int)bp->b_bcount);
      }
   
      /* Flag that we've finished */
      if (bp->b_flags & B_DONE) {
         printf ("B016Vic%d: dup biodone: link = %d\n", ctlr, link);
      }
      else
         iodone (bp);
         
      /* Note: we must NOT call 'bp_mapout' here!
       *       The unmapping is done by the call to
       *       'iodone', which looks at the B_REMAPPED
       *       bit in the 'b_flags' field of the buffer.
       */
      break;
   
   default:
      printf ("B016Vic%d: spurious interrupt\n", ctlr);
      break;
   }
   
   /* Acknowledge that we've seen the interrupt */
   r->vic[3].reg = E_NULL;
   r->vic[2].reg = 0xff;
   r->vic[1].reg = NO_LINK;
   
   return (1);
}


/* Bo16Ioctl --- handle IOCTL calls */

int B016Ioctl (dev, cmd, data, flag)
dev_t dev;     /* Major and minor device numbers */
int cmd;       /* Action code                    */
caddr_t data;  /* Address of user's buffer       */
int flag;      /* Flag word                      */
{
   int ctlr;
   int link;
   
   ctlr = B016CTLR (dev);
   link = B016LINK (dev);
   
#ifdef DB
   printf ("Ioctl: ctlr = %d, link = %d, cmd = %d\n",
                          ctlr,      link,     cmd);
#endif

   if (ctlr >= MAX_BOARDS)    /* Validate controller number */
      return (ENXIO);

   if (B016Board[ctlr].Link[link].Open == NO)   /* Is the device open ? */
      return (ENODEV);
      
   switch (cmd) {
   case READFLAGS:
      return (ReadFlags (data, ctlr, link));
      break;
   
   case SETFLAGS:
      return (SetFlags (data, ctlr, link));
      break;
   }

   return (EINVAL);     /* Invalid 'ioctl' command */
}


/* ReadFlags --- handle the ioctl READFLAGS command */

static int ReadFlags (data, ctlr, link)
caddr_t data;
int ctlr;
int link;
{
   long int *statp;
   register struct LinkState *lp;

   lp = &B016Board[ctlr].Link[link];
   statp = (long int *)data;
   *statp = 0L;
      
   if (B016Board[ctlr].Link[link].ErrFlag)
      *statp |= (long int)(BIT_0);

   return (0);                /* Success code */
}


/* SetFlags --- handle the ioctl SETFLAGS command */

static int SetFlags (data, ctlr, link)
caddr_t data;
int ctlr;
int link;
{
   long int args;
   int action;
   int param;
   int ret;
   register struct LinkState *lp;

   args = *(long int *)data;
   action = (int)(args >> 16);
   param =  (int)(args & 0xffff);
   ret = 0;                   /* Success code */
   lp = &B016Board[ctlr].Link[link];
   
   switch  (action) {
   case RESET:
      if (SendCmd (ctlr, link, RESET_CMD))
         sleep ((caddr_t)lp, B016PRI);
      else
         ret = EIO;

      break;

   case ANALYSE:
      if (SendCmd (ctlr, link, ANALYSE_CMD))
         sleep ((caddr_t)lp, B016PRI);
      else
         ret = EIO;

      break;

   case SETTIMEOUT:
      if (param == 0) {
         lp->Timed = NO;         /* No timeouts needed */
         lp->Chan[WRITE_CHAN].Timeout = DEFAULT_TIMEOUT;
         lp->Chan[READ_CHAN].Timeout = DEFAULT_TIMEOUT;
      }
      else {
         lp->Timed = YES;        /* Set up a timeout */
         lp->Chan[WRITE_CHAN].Timeout = param;
         lp->Chan[READ_CHAN].Timeout = param;
      }

      break;
      
   case SETERRORSIGNAL:
      lp->Signal = YES;
      lp->Signo  = param;
      /* Enable ERROR interrupts */
      lp->UsrProc = u.u_procp;    /* Point to user's process */
      ret = EIO;
      break;
      
   case RESETERRORSIGNAL:
      lp->Signal = NO;
      lp->Signo  = 0;
      /* Disable ERROR interrupts */
      ret = EIO;
      break;
      
   default:
      ret = EINVAL;
      break;
   }
      
#ifdef DB
   if (ret == EIO)
      printf ("SetFlags: ret = EIO\n");

   if (ret == EINVAL)
      printf ("SetFlags: ret = EINVAL\n");
#endif

   return (ret);
}


/* SendCmd --- send a command to the B016 */

static int SendCmd (ctlr, link, cmd)
int ctlr;
int link;
int cmd;
{
   register struct B016Reg *r;
   
#ifdef DB
   printf ("SendCmd: B016 %d link %d cmd %d\n", ctlr, link, cmd);
#endif

   r = B016Board[ctlr].Reg;
   
   r->vic[0].reg = cmd;
   r->ICMS[link].set = 0xff;        /* Start interrupt */

   CDELAY (r->vic[0].reg == NO_CMD, 10);

   r->ICMS[link].clr = 0xff;        /* Stop interrupt */
   
   CDELAY (r->vic[0].reg == NO_CMD, 100000);
      
   if (r->vic[0].reg == NO_CMD) {
      return (1);    /* Success */
   }
   else {
      printf ("B016Vic%d: SendCmd: state machine stuck\n", ctlr);
      return (0);    /* Failure */
   }
}


/* InitLink --- initialise a link state structure */

static void InitLink (ctlr, link)
int ctlr;
int link;
{
   B016Board[ctlr].Link[link].Open     = NO;             /* Mark the link as closed   */
   B016Board[ctlr].Link[link].ErrCode  = E_NULL;
   B016Board[ctlr].Link[link].ErrFlag  = NO;

   B016Board[ctlr].Link[link].Timed    = NO;             /* No timeouts               */

   B016Board[ctlr].Link[link].Signal   = NO;             /* No signals                */
   B016Board[ctlr].Link[link].Signo    = 0;
   B016Board[ctlr].Link[link].UsrProc  = NULL;

   InitChan (ctlr, link, READ_CHAN);
   InitChan (ctlr, link, WRITE_CHAN);
}


/* InitChan --- initialise a channel state structure */

static void InitChan (ctlr, link, chan)
int ctlr;
int link;
int chan;
{
   int i;
   char *addr;
   struct ChanState *cp;
   
   i = (link * 2) + chan;
   addr = B016Board[ctlr].SharedRam;
   cp = &B016Board[ctlr].Link[link].Chan[chan];
   
   cp->ErrCode = 0;
   cp->Timeout = DEFAULT_TIMEOUT;

   cp->B016Buf = addr + MEG + (i * MAX_BUF);
   cp->B016Len = (long int *)(addr + 0xc0000 + (4 * i));
   cp->B016Time = (long int *)(addr + 0xc0020 + (4 * i));
}


#ifdef DB2

static void DumpState (ctlr, link)
int ctlr;
int link;
{
   struct BoardState *bp;
   struct LinkState *lp;
   
   bp = &B016Board[ctlr];
   lp = &(bp->Link[link]);
   
   printf ("Controller %d:\n", ctlr);
   
   printf ("Busy      = %s\n",   bp->Busy ? "YES" : "NO");
   printf ("MemAddr   = 0x%x\n", bp->MemAddr);
   printf ("SharedRam = 0x%x\n", bp->SharedRam);
   printf ("Reg       = 0x%x\n", bp->Reg);

   printf ("Link %d:\n", link);

   printf ("Open      = %s\n",   lp->Open ? "YES" : "NO");
   printf ("ErrFlag   = %s\n",   lp->ErrFlag ? "YES" : "NO");
   printf ("Timed     = %s\n",   lp->Timed ? "YES" : "NO");
   printf ("Signal    = %s\n",   lp->Signal ? "YES" : "NO");
   printf ("Signo     = %d\n",   lp->Signo);
   printf ("UsrProc   = 0x%x\n", lp->UsrProc);
}
#endif
